import unittest
import zipfile

from tira.io_utils import zip_dir

from .format_check import IR_METADATA_MULTIPLE_VALID


class TestIoUtils(unittest.TestCase):
    def test_zipping_of_directory_01(self):
        expected = [
            "metadata.yaml",
            ".ir-metadata.yml",
            "foo-1/m.yaml",
            "foo-1/foo-metadata.yml",
            "foo-1/foo-2/m.yaml",
            "foo-1/foo-2/foo-metadata.yml",
        ]
        zipped = zip_dir(IR_METADATA_MULTIPLE_VALID)
        actual = zipfile.ZipFile(zipped).namelist()
        self.assertEqual(sorted(expected), sorted(actual))

    def test_zipping_of_directory_02(self):
        expected = ["foo-metadata.yml"]
        zipped = zip_dir(IR_METADATA_MULTIPLE_VALID / "foo-1" / "foo-2" / "foo-metadata.yml")
        actual = zipfile.ZipFile(zipped).namelist()
        self.assertEqual(sorted(expected), sorted(actual))

    def test_zipping_of_directory_03(self):
        expected = ["m.yaml"]
        zipped = zip_dir(IR_METADATA_MULTIPLE_VALID / "foo-1" / "foo-2" / "m.yaml")
        actual = zipfile.ZipFile(zipped).namelist()
        self.assertEqual(sorted(expected), sorted(actual))

    def test_zipping_of_directory_04(self):
        expected = ["metadata.yaml"]
        zipped = zip_dir(IR_METADATA_MULTIPLE_VALID / "metadata.yaml")
        actual = zipfile.ZipFile(zipped).namelist()
        self.assertEqual(sorted(expected), sorted(actual))



import os
import sys
import unittest

from tira.io_utils import MonitoredExecution


class TestRunMonitoredCmd(unittest.TestCase):
    def test_command_without_exception(self):
        def my_command(_):
            print("hello stdout 1", file=sys.stdout)
            print("hello stderr 1", file=sys.stderr)
            print("hello stdout 2", file=sys.stdout)

        actual = MonitoredExecution().run(my_command)
        self.assertTrue((actual / "output").exists())
        self.assertEqual(0, len(os.listdir(actual / "output")))
        stdout = (actual / "stdout.txt").read_text()
        stderr = (actual / "stderr.txt").read_text()

        self.assertIn("hello stdout 2", stdout)
        self.assertIn("hello stderr 1", stderr)

    def test_command_with_exception(self):
        def my_command(_):
            print("stdout 1", file=sys.stdout)
            print("stderr 1", file=sys.stderr)
            raise ValueError("foo")

        actual = MonitoredExecution().run(my_command)
        self.assertTrue((actual / "output").exists())
        self.assertEqual(0, len(os.listdir(actual / "output")))
        stdout = (actual / "stdout.txt").read_text()
        stderr = (actual / "stderr.txt").read_text()

        self.assertIn("stdout 1", stdout)
        self.assertIn("stderr 1", stderr)



import unittest

from tira.check_format import check_format, lines_if_valid

from . import (
    _ERROR,
    _OK,
    EMPTY_OUTPUT,
    IR_QUERY_OUTPUT,
    RUN_OUTPUT_WITH_DUPLICATE_DOCUMENTS,
    RUN_OUTPUT_WITH_TOO_FEW_COLUMNS,
    RUN_OUTPUT_WITH_TOO_FEW_QUERIES,
    VALID_RUN_OUTPUT,
)


class TestCheckFormatForNonExistingFormats(unittest.TestCase):
    def test_invalid_validator_on_empty_output(self):
        expected = [_ERROR, "No file run.txt or run.txt.gz was found. Only the files ['.gitkeep'] were available."]
        actual = check_format(EMPTY_OUTPUT, "run.txt")
        self.assertEqual(expected, actual)

    def test_invalid_validator_on_query_output(self):
        expected = [_ERROR, "No file run.txt or run.txt.gz was found. Only the files ['queries.jsonl'] were available."]
        actual = check_format(IR_QUERY_OUTPUT, "run.txt")
        self.assertEqual(expected, actual)

    def test_on_valid_run_output_output(self):
        expected = [_OK, "The run.txt file has the correct format."]
        actual = check_format(VALID_RUN_OUTPUT, "run.txt")
        self.assertEqual(expected, actual)

    def test_lines_on_valid_run_output(self):
        expected = 10
        actual = lines_if_valid(VALID_RUN_OUTPUT, "run.txt")
        self.assertEqual(expected, len(actual))

    def test_lines_on_invalid_run_output(self):
        with self.assertRaises(Exception):
            lines_if_valid(RUN_OUTPUT_WITH_TOO_FEW_QUERIES, "run.txt")

    def test_on_invalid_run_with_too_few_queries(self):
        expected = [_ERROR, "The run file has only 1 queries which is likely an error."]
        actual = check_format(RUN_OUTPUT_WITH_TOO_FEW_QUERIES, "run.txt")
        self.assertEqual(expected, actual)

    def test_on_invalid_run_with_duplicate_documents(self):
        expected = [
            _ERROR,
            'The run file has duplicate documents: the document with id "doc-1" appears multiple times for query "5".',
        ]
        actual = check_format(RUN_OUTPUT_WITH_DUPLICATE_DOCUMENTS, "run.txt")
        self.assertEqual(expected, actual)

    def test_on_invalid_run_with_too_short_column(self):
        expected = [
            _ERROR,
            'Invalid line in the run file, expected 6 columns, but found a line "5 doc-1 10 tag" with 4 columns.',
        ]
        actual = check_format(RUN_OUTPUT_WITH_TOO_FEW_COLUMNS, "run.txt")
        self.assertEqual(expected, actual)



from typing import List, Tuple

PANDAS_DTYPES = {
    "docno": str,
    "doc_id": str,
    "id": str,
    "qid": str,
    "query_id": str,
    "queryid": str,
    "docno": str,
    "doc_id": str,
    "docid": str,
}


class PandasIntegration:
    """Handling of inputs/outputs in TIRA with pandas. All methods here work in the TIRA sandbox without internet
    connection (when the data is mounted read-only)."""

    def __init__(self, tira_client):
        self.tira_client = tira_client

    def from_retriever_submission(
        self,
        approach: str,
        dataset: str,
        previous_stage: str = None,
        datasets: List[str] = None,
        for_ir_measures: bool = False,
    ):
        """Load a run file as pandas dataframe from tira. Compatible with PyTerrier.

        Args:
            approach (str): the approach for which the run should be loaded, in the format 'task/team/software'.
            dataset (str): the dataset id, either an tira or ir_datasets id.
            previous_stage (str, optional): The previous stage, in case the approach itself is ambigious. Defaults to
                None.
            datasets (List[str], optional): list of datasets to concat. Defaults to None.
            for_ir_measures (bool, optional): rename columns qid and docid to query_id respectively doc_id for direct
                re-use within ir-measures.

        Returns:
            pd.DataFrame: The run file parsed to a pandas DataFrame.
        """
        import pandas as pd

        from tira.ir_datasets_util import translate_irds_id_to_tirex

        task, team, software = approach.split("/")

        if dataset and datasets:
            raise ValueError(
                f"You can not pass both, dataset and datasets. Got dataset = {dataset} and datasets= {datasets}"
            )

        if not datasets:
            datasets = [dataset]

        df_ret = []
        for dataset in datasets:
            ret, run_id = self.tira_client.download_run(
                task, translate_irds_id_to_tirex(dataset), software, team, previous_stage, return_metadata=True
            )
            ret["qid"] = ret["query"].astype(str)
            ret["docno"] = ret["docid"].astype(str)
            del ret["query"]
            del ret["docid"]

            if for_ir_measures:
                ret["query_id"] = ret["qid"]
                ret["doc_id"] = ret["docno"]
                del ret["qid"]
                del ret["docno"]

            ret["tira_task"] = task
            ret["tira_dataset"] = dataset
            ret["tira_first_stage_run_id"] = run_id
            df_ret += [ret]

        return pd.concat(df_ret)

    def transform_queries(self, approach: str, dataset: str, format: str = "query-processor"):
        """Load and transform the query processing outputs specified by the approach on the dataset for direct re-use
        as a PyTerrier query transformation.

        Args:
            approach str: the approach for which the run should be loaded, in the format 'task/team/software'.
            dataset (str):the dataset id, either an tira or ir_datasets id.
            format (str, optional):  The format in which the queries are loaded.

        Raises:
            ValueError: If no approach with the identifier 'approach' was found or if there was an error parsing the
            outputs.

        Returns:
            pd.DataFrame: a DataFrame with the parsed query processing outputs compatible with PyTerrier query
            transformations.
        """
        import pandas as pd

        items = self.tira_client.iter_run_output(approach, dataset, format)

        return pd.DataFrame(items)

    def transform_documents(self, approach, dataset, format="document-processor"):
        """Load and transform the document processing outputs specified by the approach on the dataset for direct
        re-use as a PyTerrier document transformation.

        Args:
            approach str: the approach for which the run should be loaded, in the format 'task/team/software'.
            dataset (str):the dataset id, either an tira or ir_datasets id.
            format (str, optional): The format in which the documents are available.

        Raises:
            ValueError: If no approach with the identifier 'approach' was found or if there was an error parsing the
            outputs.

        Returns:
            pd.DataFrame: a DataFrame with the parsed document processing outputs compatible with PyTerrier document
            transformations.
        """
        import pandas as pd

        items = self.tira_client.iter_run_output(approach, dataset, format)

        return pd.DataFrame(items)

    def inputs(self, task, dataset=None, formats=("*.jsonl")):
        """Load the inputs to systems for a task from tira.

        Args:
            approach str: the approach for which the run should be loaded, in the format 'task/team/software'.
            dataset (str):the dataset id, either an tira or ir_datasets id.
            file_selection (Tuple[str,...], optional): The search glob to outputs specified by the approach on the
                dataset. Defaults to ('/*.jsonl', '/*.jsonl.gz').
            dtype (Tuple[str,...], optional): Transformations of the data types while loading with pandas. Defaults to
                PANDAS_DTYPES.

        Raises:
            ValueError: If the dataset is not public or does not exist.

        Returns:
            pd.DataFrame: A DataFrame with all inputs to systems.
        """
        import pandas as pd

        dataset_items = self.tira_client.iter_dataset(task, dataset, False, formats)

        return pd.DataFrame(dataset_items)

    def truths(self, task, dataset=None, formats=("*.jsonl")):
        """Load the truths, i.e., ground truth labels, for a task from tira.

        Args:
            approach str: the approach for which the run should be loaded, in the format 'task/team/software'.
            dataset (str):the dataset id, either an tira or ir_datasets id.
            file_selection (Tuple[str,...], optional): The search glob to outputs specified by the approach on the
                dataset. Defaults to ('/*.jsonl', '/*.jsonl.gz').
            dtype (Tuple[str,...], optional): Transformations of the data types while loading with pandas. Defaults to
                PANDAS_DTYPES.

        Raises:
            ValueError: If the truth is not public or does not exist.

        Returns:
            pd.DataFrame: A DataFrame with all the ground truth labels.
        """
        import pandas as pd

        dataset_items = self.tira_client.iter_dataset(task, dataset, True, formats)

        return pd.DataFrame(dataset_items)



def print_license_agreement(url):
    if not url:
        return
    if (
        url
        and "longeval" in url
        or "leipzig-topics-small-20240119" in url
        or "jena-topics-20231026-test" in url
        or "jena-topics-small-20240119-training" in url
        or "leipzig-topics-20231025-test" in url
        or "training-20231104-training" in url
        or "validation-20231104-training" in url
    ):
        print(
            'The download is derived from The LongEval Dataset under the "Qwant LongEval'
            ' Attribution-NonCommercial-ShareAlike License". Hence, the download is also under this License. By using'
            " it, you agree to the terms of this license. Please find details at:"
            " https://lindat.mff.cuni.cz/repository/xmlui/page/Qwant_LongEval_BY-NC-SA_License"
        )



import unittest

from . import (
    _ERROR,
    _OK,
    EMPTY_OUTPUT,
    TEXT_ALIGNMENT_CORPUS_INVALID,
    TEXT_ALIGNMENT_CORPUS_VALID_1,
    TEXT_ALIGNMENT_CORPUS_VALID_2,
)


def check_format(dir):
    from tira.check_format import check_format as cf

    return cf(dir, "text-alignment-corpus")


def lines_if_valid(dir):
    from tira.check_format import lines_if_valid as lf

    return lf(dir, "text-alignment-corpus")


class TestCheckJsonlForNonExistingFormats(unittest.TestCase):
    def test_invalid_validator_on_empty_output(self):
        expected = [_ERROR, "No unique pair file was found, only the files ['.gitkeep'] were available."]
        actual = check_format(EMPTY_OUTPUT)
        self.assertEqual(expected, actual)

    def test_invalid_text_alignment_corpus(self):
        expected = [
            _ERROR,
            "No unique text file was found for id suspicious-document00218.txt. Only ['pairs'] were available.",
        ]
        actual = check_format(TEXT_ALIGNMENT_CORPUS_INVALID)
        self.assertEqual(expected, actual)

    def test_valid_text_alignment_corpus_01(self):
        expected = [_OK, "The directory has the correct format."]
        actual = check_format(TEXT_ALIGNMENT_CORPUS_VALID_1)
        self.assertEqual(expected, actual)

    def test_valid_text_alignment_corpus_02(self):
        expected = [_OK, "The directory has the correct format."]
        actual = check_format(TEXT_ALIGNMENT_CORPUS_VALID_2)
        self.assertEqual(expected, actual)

    def test_lines_for_valid_corpus(self):
        expected = [
            {
                "suspicious_document_id": "suspicious-document00218",
                "suspicious_document_text": "suspicious-document00218\n",
                "source_document_id": "source-document00218",
                "source_document_text": "source-document00218\n",
            },
            {
                "suspicious_document_id": "suspicious-document00487",
                "suspicious_document_text": "suspicious-document00487\n",
                "source_document_id": "source-document00487",
                "source_document_text": "source-document00487\n",
            },
            {
                "suspicious_document_id": "suspicious-document00185",
                "suspicious_document_text": "suspicious-document00185\n",
                "source_document_id": "source-document00185",
                "source_document_text": "source-document00185\n",
            },
        ]
        actual = lines_if_valid(TEXT_ALIGNMENT_CORPUS_VALID_1)
        self.assertEqual(expected, actual)



import unittest

from tira.check_format import check_format, lines_if_valid

from . import _ERROR, _OK, EMPTY_OUTPUT, STYLE_CHANGE_CORPUS_VALID, STYLE_CHANGE_PREDICTIONS_VALID


class TestPan24StyleChangeDetectionFormats(unittest.TestCase):
    def test_invalid_validator_on_empty_output(self):
        expected = [_ERROR, "The style-change-detection directory contains only 0 instances, this is likely an error."]
        actual = check_format(EMPTY_OUTPUT, "style-change-detection-corpus")
        self.assertEqual(expected, actual)

    def test_valid_style_change_detection_corpus(self):
        expected = [_OK, "The directory has the correct format."]
        actual = check_format(STYLE_CHANGE_CORPUS_VALID, "style-change-detection-corpus")
        self.assertEqual(expected, actual)

    def test_lines_of_style_change_detection_corpus(self):
        expected = [
            {"id": "problem-1", "text": "problem 1 text\n"},
            {"id": "problem-2", "text": "problem-2 text\n"},
            {"id": "problem-3", "text": "problem-3 text\n"},
        ]
        actual = lines_if_valid(STYLE_CHANGE_CORPUS_VALID, "style-change-detection-corpus")

        self.assertEqual(expected, actual)

    def test_invalid_prediction_validator_on_empty_output(self):
        expected = [_ERROR, "The style-change-detection directory contains only 0 instances, this is likely an error."]
        actual = check_format(EMPTY_OUTPUT, "style-change-detection-predictions")
        self.assertEqual(expected, actual)

    def test_invalid_prediction_validator_on_corpus(self):
        expected = [_ERROR, "The style-change-detection directory contains only 0 instances, this is likely an error."]
        actual = check_format(STYLE_CHANGE_CORPUS_VALID, "style-change-detection-predictions")
        self.assertEqual(expected, actual)

    def test_valid_style_change_detection_predictions(self):
        expected = [_OK, "The directory has the correct format."]
        actual = check_format(STYLE_CHANGE_PREDICTIONS_VALID, "style-change-detection-predictions")
        self.assertEqual(expected, actual)

    def test_lines_style_change_detection_predictions(self):
        expected = [
            {"authors": 2, "changes": [0, 0, 1, 0, 1], "problem": "1", "type": "solution"},
            {"authors": 2, "changes": [0, 0, 1, 0, 1], "problem": "1", "type": "truth"},
            {"authors": 2, "changes": [0, 0, 1, 0, 1], "problem": "2", "type": "solution"},
        ]
        actual = lines_if_valid(STYLE_CHANGE_PREDICTIONS_VALID, "style-change-detection-predictions")
        self.assertEqual(expected, actual)



import gzip
import tempfile
from pathlib import Path
from shutil import copyfile

DF_EXPECTED_COLUMNS = ["qid", "q0", "docid", "rank", "score", "system"]


class RunSerializer:
    """Serializes a run object into a directory."""

    def is_dataframe(self, data: any):
        try:
            import pandas as pd

            return isinstance(data, pd.DataFrame)
        except:
            return False

    def serialize(self, data, output_path: Path):
        output_path.mkdir(exist_ok=True, parents=True)

        if isinstance(data, str):
            return self.serialize(Path(data), output_path)
        if isinstance(data, Path):
            if not data.is_file():
                raise ValueError(f"The passed run file {data} is not a file.")
            copyfile(data, output_path / ("run.txt.gz" if (str(data).endswith(".gz")) else "run.txt"))
            return
        if self.is_dataframe(data):
            data = data.copy()
            orig_columns = [i for i in data.columns]
            if "q0" not in data.columns:
                data["q0"] = "Q0"

            if "qid" not in data.columns and "query" in data.columns:
                data["qid"] = data["query"]

            if "docno" in data.columns:
                data["docid"] = data["docno"]

            missing_columns = []
            for f in DF_EXPECTED_COLUMNS:
                if f not in data.columns:
                    missing_columns += [f]

            if len(missing_columns) > 0:
                raise ValueError(
                    f"The passed run is missing the required columns {missing_columns}. It has the columns {orig_columns}."
                )

            lines = []
            for _, i in data.iterrows():
                l = ""
                for f in DF_EXPECTED_COLUMNS:
                    if len(str(i[f]).strip().split()) > 1:
                        raise ValueError(
                            f'Fields in a run file are expected to not contain whitespaces. The value "{str(i[f]).strip()}" for the field "{f}" contains whitespaces.'
                        )
                    l += " " + str(i[f])
                lines += [l.strip()]

            with gzip.open(output_path / "run.txt.gz", "wt") as f:
                f.write("\n".join(lines))
            return
        if hasattr(data, "run_data") and self.is_dataframe(data.run_data):
            return self.serialize(data.run_data, output_path)

        raise ValueError("foo")


FORMAT_TO_SERIALIZER = {"run.txt": lambda: RunSerializer()}

SUPPORTED_FORMATS = set(sorted(list(FORMAT_TO_SERIALIZER.keys())))


def serialize_in_temporary_directory(data: any, format: str):
    """Check if the provided run output is in the specified format. Provides debug messages intended for users.

    Args:
        format (Union[str, Sequence[str]]): The allowed format or a list of allowed formats.
        run_output (Path): the output produced by some run that is to-be checked.
    """
    if format not in SUPPORTED_FORMATS:
        raise ValueError(f"Format {format} is not supported. Supported formats are {SUPPORTED_FORMATS}.")
    ret = tempfile.TemporaryDirectory()
    ret = Path(ret.name)
    serializer = FORMAT_TO_SERIALIZER[format]()
    serializer.serialize(data, ret)
    return ret



import copy
import gzip
import json
import os
from base64 import b64encode
from pathlib import Path
from typing import Any, Iterable, Optional

import pandas as pd
from bs4 import BeautifulSoup
from tqdm import tqdm


class IrDatasetsLoader(object):
    """Base class for loading datasets in a standardized format"""

    def load_irds(self, ir_datasets_id):
        from tira.third_party_integrations import ir_datasets

        try:
            return ir_datasets.load(ir_datasets_id)
        except Exception:
            raise ValueError(f"Could not load the dataset {ir_datasets_id}. Does it exist?")

    def yield_docs(self, dataset, include_original, skip_duplicate_ids, allowlist_path_ids):
        already_covered_ids = set()
        allowed_ids = set()
        if allowlist_path_ids:
            with open(allowlist_path_ids, "r") as inp_file:
                for i in inp_file:
                    allowed_ids.add(i.strip())
            print("I use a allow list of size ", len(allowed_ids))

        for doc in tqdm(dataset.docs_iter(), "Load Documents"):
            if skip_duplicate_ids and doc.doc_id in already_covered_ids:
                continue
            if allowlist_path_ids and str(doc.doc_id) not in allowed_ids:
                continue

            yield self.map_doc(doc, include_original)
            if skip_duplicate_ids:
                already_covered_ids.add(doc.doc_id)

    def load_dataset_for_fullrank(
        self,
        ir_datasets_id: str,
        output_dataset_path: Path,
        output_dataset_truth_path: Path,
        include_original=True,
        skip_documents=False,
        skip_qrels=False,
        skip_duplicate_ids=True,
        allowlist_path_ids: "Optional[Path]" = None,
    ) -> None:
        """Loads a dataset through the ir_datasets package by the given ir_datasets ID.
        Maps documents, queries, qrels to a standardized format in preparation for full-rank operations with PyTerrier.

        @param ir_datasets_id: the dataset ID as of ir_datasets
        @param output_dataset_path: the path to the directory where the output files will be stored
        @param output_dataset_truth_path: the path to the directory where the output files will be stored
        @param include_original {False}: flag which signals if the original data of documents and queries should be included
        @param skip_duplicate_ids: Should this pipeline skip duplicate ids?
        @param allowlist_path_ids: skip ids not in the allowlist (e.g., for filtering the subcategories of the ClueWebs)
        """
        dataset = self.load_irds(ir_datasets_id)

        if not skip_documents and output_dataset_path:
            self.write_lines_to_file(
                self.yield_docs(dataset, include_original, skip_duplicate_ids, allowlist_path_ids),
                output_dataset_path / "documents.jsonl",
            )

        queries_mapped_jsonl = [self.map_query_as_jsonl(query, include_original) for query in dataset.queries_iter()]
        queries_mapped_xml = [self.map_query_as_xml(query, include_original) for query in dataset.queries_iter()]

        if not skip_qrels:
            try:
                qrels_mapped = [self.map_qrel(qrel) for qrel in dataset.qrels_iter()]
            except Exception:
                print(
                    'WARNING: I could not load qrels and will skip writing the file "qrels.txt". This is expected if'
                    " your dataset has no qrels yet. If you have qrels, please debug this problem locally on your"
                    " machine."
                )
                qrels_mapped = []

            if len(qrels_mapped) > 0:
                self.write_lines_to_file(qrels_mapped, output_dataset_truth_path / "qrels.txt")

        if output_dataset_path:
            self.write_lines_to_file(queries_mapped_jsonl, output_dataset_path / "queries.jsonl")
            self.write_lines_to_file(
                [json.dumps({"ir_datasets_id": ir_datasets_id})], output_dataset_path / "metadata.json"
            )
            self.write_lines_to_xml_file(ir_datasets_id, queries_mapped_xml, output_dataset_path / "queries.xml")

        if output_dataset_truth_path:
            self.write_lines_to_file(queries_mapped_jsonl, output_dataset_truth_path / "queries.jsonl")
            self.write_lines_to_xml_file(ir_datasets_id, queries_mapped_xml, output_dataset_truth_path / "queries.xml")

    def reformat_to_re_rank_dataset(
        self, run_file: Path, dataset_directory: Path, output_dataset_dir: Path, log_file: Path
    ):
        run_file = Path(run_file).absolute()
        dataset_directory = Path(dataset_directory).absolute()

        if not log_file.exists():
            with open(log_file, "w") as f:
                f.write("")

        with open(log_file, "r") as f:
            for l in f:
                try:
                    l = json.loads(l)
                    if (
                        l["dataset_directory"] == str(dataset_directory)
                        and l["re_rank_file"]
                        and l["run_file"] == str(run_file)
                    ):
                        return str(Path(l["re_rank_dataset_directory"]) / "input-data")
                except:
                    pass

        import ir_datasets

        from tira.ir_datasets_util import static_ir_dataset

        dataset = static_ir_dataset(str(dataset_directory)).load("does-not-matter")
        ir_datasets.registry.register(dataset.dataset_id, dataset)

        self.load_dataset_for_rerank(
            dataset.dataset_id,
            output_dataset_path=output_dataset_dir / "input-data",
            output_dataset_truth_path=None,
            include_original=False,
            run_file=run_file,
        )

        ret = {
            "dataset_directory": str(dataset_directory),
            "re_rank_file": True,
            "run_file": str(run_file),
            "re_rank_dataset_directory": str(output_dataset_dir),
        }

        with open(log_file, "a") as f:
            f.write(json.dumps(ret) + "\n")

        return Path(ret["re_rank_dataset_directory"]) / "input-data"

    def load_dataset_for_rerank(
        self,
        ir_datasets_id: str,
        output_dataset_path: Path,
        output_dataset_truth_path: Path,
        include_original: bool,
        run_file: Path,
    ) -> None:
        """Loads a dataset through ir_datasets package by the given ir_datasets ID.
        Maps qrels and TREC-run-formatted data by a given file to a format fitted for re-rank operations with PyTerrier.

        @param ir_datasets_id: the dataset ID as of ir_datasets
        @param output_dataset_path: the path to the directory where the output files will be stored
        @param output_dataset_truth_path: the path to the directory where the output files will be stored
        @param include_original {False}: flag which signals if the original data of documents and queries should be included
        @param run_file: the path to a file with data in TREC-run format
        """
        dataset = self.load_irds(ir_datasets_id)
        queries = {str(i.query_id): i for i in dataset.queries_iter()}

        run = self.load_run_file(run_file)
        print("Get Documents")
        docs = self.get_docs_by_ids(dataset, set(str(i["docno"]) for i in run))
        print("Produce rerank data.")
        rerank = tqdm(
            (self.construct_rerank_row(docs, queries, i, include_original) for i in run), "Produce Rerank File."
        )
        print("Write rerank data.")
        self.write_lines_to_file(rerank, output_dataset_path / "rerank.jsonl.gz")
        print("Done rerank data was written.")
        if output_dataset_truth_path:
            print("Write qrels data.")
            qrels_mapped = (self.map_qrel(qrel) for qrel in dataset.qrels_iter())
            self.write_lines_to_file(qrels_mapped, output_dataset_truth_path / "qrels.txt")

    def map_doc(self, doc: tuple, include_original=True) -> str:
        """Maps a document of any dataset (loaded through ir_datasets) to a standarized format
        stores full document data too, if flag 'include_original' is set

        @param doc: the document as a namedtuple
        @param include_original: flag which signals if the original document data should be stored too
        :return ret: the mapped document
        """
        ret = {"docno": doc.doc_id, "text": doc.default_text()}
        if include_original:
            ret["original_document"] = self.make_serializable(doc._asdict())
        return json.dumps(ret)

    def map_query_as_jsonl(self, query: tuple, include_original=True) -> str:
        ret = {
            "qid": query.query_id,
            "query": query.default_text(),
        }
        if include_original:
            ret["original_query"] = query._asdict()
        return json.dumps(ret)

    def map_query_as_xml(self, query: tuple, include_original=False) -> str:
        soup = BeautifulSoup()
        soup.append(soup.new_tag("topic", attrs={"number": query.query_id}))
        soup.topic.append(soup.new_tag("query"))
        soup.query.append(soup.new_string(query.default_text()))

        if include_original:
            soup.topic.append(soup.new_tag("original_query"))
            for key, value in query._asdict().items():
                soup.original_query.append(soup.new_tag(str(key)))
                tag = soup.original_query.find(key)
                tag.append(soup.new_string(str(value)))
        return soup

    def map_qrel(self, qrel: tuple) -> str:
        return f"{qrel.query_id} {qrel.iteration} {qrel.doc_id} {qrel.relevance}"

    def load_run_file(self, run_file: Path) -> list:
        if not os.path.abspath(run_file).endswith("run.txt"):
            run_file = run_file / "run.txt"

        run = pd.read_csv(
            os.path.abspath(run_file), sep="\\s+", names=["qid", "Q0", "docno", "rank", "score", "system"]
        )
        run = run.copy().sort_values(["qid", "score", "docno"], ascending=[True, False, False]).reset_index()
        run = run.groupby("qid")[["qid", "Q0", "docno", "rank", "score", "system"]].head(1000)

        # Make sure that rank position starts by 1
        run["rank"] = 1
        run["rank"] = run.groupby("qid")["rank"].cumsum()

        return [i.to_dict() for _, i in run[["qid", "Q0", "docno", "rank", "score", "system"]].iterrows()]

    def get_docs_by_ids(self, dataset, doc_ids: set[str]) -> dict[Any, Any]:
        docstore = dataset.docs_store()
        try:
            ret = {}
            doc_ids = set(doc_ids)
            for doc in tqdm(docstore.get_many_iter(doc_ids), total=len(doc_ids), desc="Get Docs"):
                ret[doc.doc_id] = doc
            return ret
        except Exception:
            ret = {}
            doc_ids = set(doc_ids)
            for doc_id in tqdm(doc_ids, total=len(doc_ids), desc="Get Docs"):
                doc = docstore.get(doc_id)
                ret[doc.doc_id] = doc
            return ret

    def make_serializable(self, o: dict):
        for k in o.keys():
            value = o[k]
            if value and value.__class__ and str(value.__class__.__name__) == "bytes":
                o[k] = b64encode(value).decode("ascii")

        return o

    def construct_rerank_row(
        self, docs: dict, queries: dict, rerank_line: "dict[str, Any]", include_original
    ) -> "Optional[str]":
        query = queries[str(rerank_line["qid"])]
        doc = docs.get(str(rerank_line["docno"]), None)

        if not doc:
            return None

        ret = {
            "qid": query.query_id,
            "query": query.default_text(),
            "original_query": self.make_serializable(query._asdict()),
            "docno": rerank_line["docno"],
            "text": doc.default_text(),
            "original_document": {},
            "rank": rerank_line["rank"],
            "score": rerank_line["score"],
        }

        if include_original:
            ret["original_document"] = (self.make_serializable(doc._asdict()),)

        return json.dumps(ret)

    def write_lines_to_file(self, lines: "Iterable[str]", path: "Path") -> None:
        if path.exists():
            raise RuntimeError(f"File already exists: {path}")
        path.parent.mkdir(parents=True, exist_ok=True)

        if os.path.abspath(path).endswith(".gz"):
            with gzip.open(os.path.abspath(path), "wb") as file:
                for line in lines:
                    if not line:
                        continue
                    file.write((line + "\n").encode("utf-8"))
        else:
            with path.open("wt") as file:
                file.writelines("%s\n" % line for line in lines if line)

    def write_lines_to_xml_file(self, ir_datasets_id: str, lines: "Iterable[str]", path: "Path") -> None:
        if path.exists():
            raise RuntimeError(f"File already exists: {path}")
        path.parent.mkdir(parents=True, exist_ok=True)
        soup = BeautifulSoup()
        soup.append(soup.new_tag("topics", attrs={"ir-datasets-id": ir_datasets_id}))
        root = soup.find("topics")
        for line in lines:
            root.append(copy.deepcopy(line))
        with path.open("wt") as file:
            file.write(soup.prettify())



import os
import uuid
from pathlib import Path

from tira.tirex import IRDS_TO_TIREX_DATASET


class PyTerrierIntegration:
    def __init__(self, tira_client):
        self.tira_client = tira_client
        self.pd = tira_client.pd
        self.irds_docker_image = "webis/tira-application:0.0.36"

    def retriever(self, approach, dataset=None):
        from tira.pyterrier_util import TiraFullRankTransformer

        input_dir = self.ensure_dataset_is_cached(dataset, dataset)
        return TiraFullRankTransformer(approach, self.tira_client, input_dir)

    def ensure_dataset_is_cached(self, irds_dataset_id, dataset):
        import os
        from pathlib import Path

        from tira.io_utils import run_cmd

        cache_dir = self.tira_client.tira_cache_dir + "/pyterrier/" + irds_dataset_id
        full_rank_data = cache_dir + "/full-rank/"
        truth_data = cache_dir + "/truth-data/"
        irds_cache = cache_dir + "/irds-cache/"

        if os.path.isfile(full_rank_data + "/documents.jsonl"):
            return full_rank_data

        Path(full_rank_data).mkdir(parents=True, exist_ok=True)
        Path(truth_data).mkdir(parents=True, exist_ok=True)
        Path(irds_cache).mkdir(parents=True, exist_ok=True)

        run_cmd(
            [
                "docker",
                "run",
                "-v",
                irds_cache + ":/root/.ir_datasets/:rw",
                "-v",
                full_rank_data + ":/output/:rw",
                "-v",
                truth_data + ":/truth/:rw",
                "--entrypoint",
                "/irds_cli.sh",
                self.irds_docker_image,
                "--output_dataset_path",
                "/output",
                "--ir_datasets_id",
                irds_dataset_id,
                "--output_dataset_truth_path",
                "/truth",
            ]
        )

        return full_rank_data

    def create_rerank_file(self, run_df=None, run_file=None, irds_dataset_id=None):
        import gzip
        import json
        import tempfile
        from pathlib import Path

        from tira.io_utils import run_cmd
        from tira.third_party_integrations import persist_and_normalize_run

        if run_df is None and run_file is None:
            raise ValueError("Please pass either run_df or run_file")

        if run_file is not None:
            return run_file

        run_file = tempfile.TemporaryDirectory("-rerank-run").name
        Path(run_file).mkdir(parents=True, exist_ok=True)

        if "text" not in run_df.columns and "body" not in run_df.columns:
            if not irds_dataset_id:
                raise ValueError(f"Please pass a irds_dataset_id. Got {irds_dataset_id}.")
            persist_and_normalize_run(run_df, "system-is-ignored", run_file)

            cache_dir = self.tira_client.tira_cache_dir + "/pyterrier/" + irds_dataset_id
            irds_cache = cache_dir + "/irds-cache/"

            run_cmd(
                [
                    "docker",
                    "run",
                    "-v",
                    irds_cache + ":/root/.ir_datasets/:rw",
                    "-v",
                    run_file + ":/output/:rw",
                    "--entrypoint",
                    "/irds_cli.sh",
                    self.irds_docker_image,
                    "--output_dataset_path",
                    "/output",
                    "--ir_datasets_id",
                    irds_dataset_id,
                    "--rerank",
                    "/output",
                ]
            )
        else:
            with gzip.open(run_file + "/rerank.jsonl.gz", "wt") as f:
                for _, i in run_df.iterrows():
                    i = i.to_dict()

                    for k in ["original_query", "original_document"]:
                        if k not in i:
                            i[k] = {}

                    if "text" not in i and "body" in i:
                        i["text"] = i["body"]

                    if "text" not in i:
                        raise ValueError(f'I expect a field "text", but only found fields {i.keys()}.')

                    f.write(json.dumps(i) + "\n")

        return run_file

    def index(self, approach, dataset):
        """
        Load an PyTerrier index from TIRA.
        """
        import pyterrier as pt

        from tira.ir_datasets_util import translate_irds_id_to_tirex

        ret = self.tira_client.get_run_output(approach, translate_irds_id_to_tirex(dataset)) + "/index"
        return pt.IndexFactory.of(os.path.abspath(ret))

    def from_submission(self, approach, dataset=None, datasets=None, file_to_re_rank=None):
        if self.__is_re_ranker(approach):
            return self.from_retriever_submission(approach, dataset, datasets=datasets, file_to_re_rank=file_to_re_rank)
        else:
            from tira.pyterrier_util import TiraRerankingTransformer

            return TiraRerankingTransformer(approach, self.tira_client, dataset, datasets)

    def __is_re_ranker(self, approach):
        if self.tira_client.input_run_in_sandbox(approach):
            # If the input run is in the sandbox, everything behaves as a re-ranker
            return True

        _, team, software = approach.split("/")
        software = self.tira_client.public_system_details(team, software)
        return software.get("ir_re_ranker", False)

    def from_retriever_submission(self, approach, dataset, previous_stage=None, datasets=None, file_to_re_rank=None):
        from tira.ir_datasets_loader import IrDatasetsLoader
        from tira.pyterrier_util import TiraSourceTransformer

        if file_to_re_rank:
            re_rank_dir = Path(self.tira_client.tira_cache_dir) / "extracted_datasets" / str(uuid.uuid4())
            log_file = Path(f"{self.tira_client.tira_cache_dir}/.archived/local-executions.jsonl")
            re_rank_dir = IrDatasetsLoader().reformat_to_re_rank_dataset(
                Path(file_to_re_rank), dataset, re_rank_dir, log_file
            )
            dataset = str(Path(re_rank_dir))

        ret = self.pd.from_retriever_submission(approach, dataset, previous_stage, datasets)
        return TiraSourceTransformer(ret)

    def transform_queries(self, approach, dataset, format: str = "query-processor", prefix: str = ""):
        from pyterrier.apply import generic

        ret = self.pd.transform_queries(approach, dataset, format)
        cols = [i for i in ret.columns if i not in ["qid"]]
        ret = {str(i["qid"]): i for _, i in ret.iterrows()}

        def __transform_df(df):
            df = df.copy()
            for col in cols:
                df[prefix + col] = df["qid"].apply(lambda i: ret[str(i)][col])
            return df

        return generic(__transform_df)

    def transform_documents(self, approach, dataset, format: str = "document-processor", prefix: str = ""):
        from pyterrier.apply import generic

        ret = self.pd.transform_documents(approach, dataset, format)
        cols = [i for i in ret.columns if i not in ["docno"]]
        ret = {str(i["docno"]): i for _, i in ret.iterrows()}

        def __transform_df(df):
            for col in cols:
                df[prefix + col] = df["docno"].apply(lambda i: ret[str(i)][col])
            return df

        return generic(__transform_df)

    @staticmethod
    def _get_features_from_row(row, cols, map_features=None):
        import numpy as np

        res = []

        for c in cols:
            if map_features is not None and c in map_features:
                f = map_features[c](row[c])
            else:
                f = row[c]

            if isinstance(f, (list, np.ndarray)):
                res.extend(f)
            else:
                res.append(f)

        return np.array(res)

    def _features_transformer(self, run, id_col, name, feature_selection=None, map_features=None):
        from tira.pyterrier_util import TiraApplyFeatureTransformer

        cols = [col for col in run.columns if col != id_col]
        if feature_selection is not None:
            cols = [col for col in cols if col in feature_selection]

        mapping = {str(row[id_col]): self._get_features_from_row(row, cols, map_features) for _, row in run.iterrows()}

        return TiraApplyFeatureTransformer(mapping, (id_col,), name)

    def doc_features(
        self, approach, dataset, format: str = "document-processor", feature_selection=None, map_features=None
    ):
        run = self.pd.transform_documents(approach, dataset, format)

        return self._features_transformer(run, "docno", "doc_features", feature_selection, map_features)

    def query_features(
        self, approach, dataset, format: str = "query-processor", feature_selection=None, map_features=None
    ):
        run = self.pd.transform_queries(approach, dataset, format)

        return self._features_transformer(run, "qid", "query_features", feature_selection, map_features)

    def reranker(self, approach, irds_id=None):
        from tira.pyterrier_util import TiraLocalExecutionRerankingTransformer

        return TiraLocalExecutionRerankingTransformer(approach, self.tira_client, irds_id=irds_id)


class PyTerrierAnceIntegration:
    """
    The pyterrier_ance integration to re-use cached ANCE indices. Wraps https://github.com/terrierteam/pyterrier_ance
    """

    def __init__(self, tira_client):
        self.tira_client = tira_client

    def ance_retrieval(self, dataset: str):
        """Load a cached pyterrier_ance.ANCEIndexer submitted as workshop-on-open-web-search/ows/pyterrier-anceindex
        from tira.

        References (for citation):
            https://arxiv.org/pdf/2007.00808.pdf
            https://github.com/microsoft/ANCE/

        Args:
            dataset (str): the dataset id, either an tira or ir_datasets id.

        Returns:
            pyterrier_ance.ANCERetrieval: the ANCE index.
        """
        from tira.ir_datasets_util import translate_irds_id_to_tirex

        ance_index = (
            Path(
                self.tira_client.get_run_output(
                    "ir-lab-sose-2024/ows/pyterrier-anceindex", translate_irds_id_to_tirex(dataset)
                )
            )
            / "anceindex"
        )
        ance_checkpoint = self.tira_client.load_resource("Passage_ANCE_FirstP_Checkpoint.zip")
        import pyterrier_ance

        return pyterrier_ance.ANCERetrieval(ance_checkpoint, ance_index)


class PyTerrierSpladeIntegration:
    """The pyt_splade integration to re-use cached Splade indices. Wraps https://github.com/cmacdonald/pyt_splade"""

    def __init__(self, tira_client):
        self.tira_client = tira_client

    def splade_index(self, dataset: str, approach: str = "workshop-on-open-web-search/naverlabseurope/Splade (Index)"):
        """
        Load a cached pyt_splade index submitted as the passed approach (default
        'workshop-on-open-web-search/naverlabseurope/Splade (Index)') from tira.

        References (for citation):
            https://github.com/naver/splade?tab=readme-ov-file#cite-scroll
            ToDo: Ask Thibault what to cite.

        Args:
            dataset (str): the dataset id, either an tira or ir_datasets id.
            approach (str, optional): the approach id, defaults
                'workshop-on-open-web-search/naverlabseurope/Splade (Index)'.

        Returns:
            The PyTerrier index suitable for retrieval.
        """
        import pyterrier as pt

        from tira.ir_datasets_util import translate_irds_id_to_tirex

        ret = (
            Path(
                self.tira_client.get_run_output(
                    "ir-lab-sose-2024/naverlabseurope/Splade (Index)", translate_irds_id_to_tirex(dataset)
                )
            )
            / "spladeindex"
        )
        return pt.IndexFactory.of(os.path.abspath(ret))


def pt_transformer(path):
    import pyterrier as pt

    if not pt.started():
        pt.init()
    # TODO hacked for the moment, in reality, we must delegate to the classes above.
    return pt.transformer.get_transformer(pt.io.read_results(path + "/output/run.txt"))



#!/usr/bin/env python
import argparse
import logging
from pathlib import Path
from platform import python_version
from typing import TYPE_CHECKING, Optional

from tira import __version__
from tira.check_format import fmt_message
from tira.io_utils import _fmt, log_message, verify_tira_installation
from tira.rest_api_client import Client as RestClient
from tira.tira_run import guess_system_details, guess_vm_id_of_user

if TYPE_CHECKING:
    from .tira_client import TiraClient


def download_run(client, approach, dataset):
    help()


"""
Utility functions to setup arguments that are common across multiple commands (e.g., logging).
"""


def setup_logging_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-v",
        "--verbose",
        default=0,
        action="count",
        help="Increases the output verbosity. Default is INFO. Can be repeated for more verbose output.",
    )
    parser.add_argument("-q", "--quiet", action="store_true", help="Disables logging.")


"""
Setup commands for each supported subcommand.
"""


def setup_download_command(parser: argparse.ArgumentParser) -> None:
    setup_logging_args(parser)
    parser.add_argument(
        "--approach",
        required=False,
        default=None,
        help="Download the outputs of the specified approach. Usage: --approach <task-id>/<user-id>/<approach-name>",
    )
    parser.add_argument("--dataset", required=True, help="The dataset.")
    parser.set_defaults(executable=download_command)


def setup_login_command(parser: argparse.ArgumentParser) -> None:
    setup_logging_args(parser)
    parser.add_argument("--token", required=True, default=None, help="The token to login to the server.")
    parser.add_argument(
        "--print-docker-auth",
        required=False,
        default=False,
        action="store_true",
        help="Print the docker credentials as TIRA_DOCKER_REGISTRY_USER and TIRA_DOCKER_REGISTRY_TOKEN to stdout.",
    )
    parser.set_defaults(executable=login_command)


def setup_evaluation_command(parser: argparse.ArgumentParser) -> None:
    setup_logging_args(parser)
    parser.add_argument(
        "--predictions", required=True, default=None, help="The directory with the predictions to evaluate."
    )
    parser.add_argument(
        "--truths", required=False, default=None, help="The optional truths (use truths from the config if available)."
    )
    parser.add_argument("--config", required=True, help="The dataset for which the predictions are made.")
    parser.set_defaults(executable=evaluate_command)


def evaluate_command(predictions: Path, truths: Path, dataset: str, **kwargs) -> int:
    client: "TiraClient" = RestClient()
    eval_config = client.get_dataset(dataset)
    if not truths:
        if "task_id" in eval_config:
            task_id = eval_config["task_id"]
        elif "default_task" in eval_config:
            task_id = eval_config["default_task"]
        else:
            raise ValueError("Task configuration is invalid")

        truths = Path(client.download_dataset(task_id, eval_config["dataset_id"], truth_dataset=True))

    from tira.evaluators import evaluate, load_evaluator_config

    use_unsandboxed_evaluator = False

    try:
        load_evaluator_config(eval_config, client)
        use_unsandboxed_evaluator = True
    except:
        pass

    if use_unsandboxed_evaluator:
        print(evaluate(Path(predictions), Path(truths), eval_config))
    else:
        client.evaluate(Path(predictions), dataset)

    return 0


def setup_code_submission_command(parser: argparse.ArgumentParser) -> None:
    setup_logging_args(parser)
    parser.add_argument(
        "--path",
        required=True,
        default=None,
        help="The path used to build the docker image, must be in a clean git repository.",
    )
    parser.add_argument(
        "--command",
        required=False,
        default=None,
        help="The command that executes your approach in the docker image (default: the Docker Entrypoint). The command should read the data from $inputDataset to write predictions to $outputDir.",
    )
    parser.add_argument(
        "--dry-run",
        required=False,
        default=False,
        action="store_true",
        help="Make a dry-run, i.e., to develop your solution, i.e., not uploading to TIRA and not requiring that the git repo is clean.",
    )
    parser.add_argument(
        "--allow-network",
        required=False,
        default=False,
        action="store_true",
        help="Allow network communication. Within TIRA itself, software is executed in a sandbox without internet access",
    )
    parser.add_argument("--task", required=True, default=None, help="The task to which the code should be submitted.")
    parser.add_argument(
        "--dataset", required=False, default=None, help="The dataset on which the code should be tested before upload."
    )

    parser.set_defaults(executable=code_submission_command)


def setup_verify_installation(parser: argparse.ArgumentParser) -> None:
    setup_logging_args(parser)
    parser.set_defaults(executable=verify_installation_command)


def setup_upload_command(parser: argparse.ArgumentParser) -> None:
    setup_logging_args(parser)
    parser.add_argument(
        "--directory", required=True, default=None, help="The directory with the predictions to upload."
    )
    parser.add_argument("--dataset", required=True, help="The dataset for to which the predictions should be uploaded.")
    parser.add_argument(
        "--dry-run",
        required=False,
        default=False,
        action="store_true",
        help="Make a dry-run, i.e., to verify that your output is valid without uploading to TIRA.",
    )
    parser.add_argument(
        "--system", required=False, default=None, help="The system name under which the run should be uploaded."
    )
    parser.set_defaults(executable=upload_command)


def setup_eval_command(parser: argparse.ArgumentParser) -> None:
    setup_logging_args(parser)
    parser.add_argument(
        "--predictions",
        required=True,
        default=None,
        help="The path that contains the predictions.",
    )

    parser.add_argument(
        "--truths",
        required=False,
        default=None,
        help="Optional: the path that contains the truths (use the dataset from the dataset otherwise).",
    )

    parser.add_argument("--dataset", required=False, help="The dataset for which the predictions are made.")
    parser.set_defaults(executable=evaluate_command)


"""
Implementations of each command. They don't all need to be here but if your command is short and sweet and you don't
don't know, where else to put it, this is a good place.
"""


def download_command(dataset: str, approach: "Optional[str]" = None, **kwargs) -> int:
    client: "TiraClient" = RestClient()
    if approach is not None:
        print(client.get_run_output(approach, dataset))
    else:
        print(client.download_dataset(None, dataset))
    return 0


def login_command(token: str, print_docker_auth: bool, **kwargs) -> int:
    client: "TiraClient" = RestClient()
    client.login(token)
    if print_docker_auth:
        print(client.local_execution.docker_client_is_authenticated())
    return 0


def code_submission_command(
    path: Path,
    task: str,
    dry_run: bool,
    allow_network: bool,
    command: "Optional[str]",
    dataset: "Optional[str]",
    **kwargs,
) -> int:
    client: "TiraClient" = RestClient()
    client.submit_code(Path(path), task, command, dry_run=dry_run, allow_network=allow_network, dataset_id=dataset)

    return 0


def verify_installation_command(**kwards) -> int:
    status = verify_tira_installation()

    print("\nResult:")
    msg = "Your TIRA installation is valid."
    if status == _fmt.WARN:
        msg = "There are some warnings, you might not use all features."

    if status == _fmt.ERROR:
        msg = "Your installation is not valid, you might not use all features."

    log_message(msg, status)
    return 0 if status == _fmt.OK else 1


def upload_command(dataset: str, directory: Path, dry_run: bool, system: str, **kwargs) -> None:
    client: "TiraClient" = RestClient()
    vm_id = None
    default_task = None
    if client.api_key_is_valid():
        system = guess_system_details(directory, system)
        dataset_info = client.get_dataset(dataset=dataset)
        default_task = dataset_info["default_task"]
        vm_id = guess_vm_id_of_user(default_task, client)
        if not system:
            print(
                fmt_message(
                    "Please specify the name of your system. Either:"
                    + "\n\n\tIncorporate the tag into your ir-metadata (see https://ir-metadata.org),"
                    + "\n\n\tor, pass --system to tira-cli upload",
                    _fmt.ERROR,
                )
            )
            return 1

    resp = client.upload_run_anonymous(directory, dataset, dry_run, verbose=not system and not vm_id)
    if not resp or "uuid" not in resp or not resp["uuid"]:
        raise ValueError("Upload failed...")

    if not system or not vm_id:
        # only anonymous submissions
        return 0
    else:
        print("\nI upload the metadata for the submission...")
        resp = client.claim_ownership(
            resp["uuid"], vm_id, system["tag"], system.get("description", "todo: Add a description"), default_task
        )
        if "status" not in resp or "0" != resp["status"]:
            raise ValueError("There was an error with the upload, please try again...")
        print(
            "\t"
            + fmt_message(
                f"Done. Your run is available at:\n\thttps://www.tira.io/submit/{default_task}/user/{vm_id}/upload-submission",
                _fmt.OK,
            )
        )
        return 0


"""
Finally, parsing the arguments and running the chose command.
"""


def parse_args():
    parser = argparse.ArgumentParser(prog="tira-cli")
    version_str = f"TIRA v{__version__} using Python v{python_version()}"
    parser.add_argument("-v", "--version", action="version", version=version_str)

    # Register all subcommands here:
    subparsers = parser.add_subparsers(dest="command", required=True)
    setup_download_command(subparsers.add_parser("download", help="Download runs or datasets from TIRA.io"))
    setup_upload_command(subparsers.add_parser("upload", help="Upload runs or datasets to TIRA.io"))
    setup_eval_command(subparsers.add_parser("evaluate", help="Evaluate runs locally."))
    setup_login_command(subparsers.add_parser("login", help="Login your TIRA client to the tira server."))
    setup_verify_installation(
        subparsers.add_parser("verify-installation", help="Verify that your local TIRA client is correctly installed.")
    )
    setup_code_submission_command(
        subparsers.add_parser("code-submission", help="Make a code submission via Docker from a git repository.")
    )

    return parser.parse_args()


def main() -> int:
    args = parse_args()
    if not hasattr(args, "executable") or args.executable is None:
        logging.critical("No subcommand was found or it was not set up correctly.")
        logging.critical(
            f"Command: {vars(args).get('command', None)}; Executable: {vars(args).get('executable', None)}"
        )
        return -1

    # Potential bug: a user could supply his own "quiet" argument without using setup_logging_args and we would handle
    # it as if it was a logging argument, which it may not be. Probably the best solution right now would be to create a
    # custom argparse.Action for setup_logging_args to handle the --quiet and --verbose options... but that would be a
    # bit more effort and we can tackle that when this actually clashes (i.e., if we need --quiet or --verbose handled
    # differently in another command.
    if hasattr(args, "quiet") and args.quiet:
        logging.disable()
    else:
        if not hasattr(args, "verbose"):
            args.verbose = 0
        verbosity_levels = [logging.CRITICAL, logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
        default_verbosity = verbosity_levels.index(logging.INFO)
        verbosity = verbosity_levels[min(default_verbosity + args.verbose, len(verbosity_levels) - 1)]
        logging.basicConfig(level=verbosity)

    # Subcommands store their executable into args.executable which takes the parsed arguments and returns an integer.
    # We just dump all the arguments into the function and let it sort it out itself.
    return args.executable(**vars(args))



import os
from abc import ABC
from collections import defaultdict
from pathlib import Path
from statistics import mean
from typing import Any, Dict, List, Optional, Union

CONF_ID_FIELD = "id_field"
from tira.check_format import (
    CONF_ID_FIELD,
    CONF_VALUE_FIELD,
    _fmt,
    check_format,
    check_format_configuration_if_valid,
    lines_if_valid,
    log_message,
)
from tira.io_utils import to_prototext
from tira.rest_api_client import Client
from tira.tira_client import TiraClient


class TiraBaseEvaluator(ABC):
    def __init__(
        self,
        run_format: "Union[str, List[str]]",
        run_format_configuration: "Optional[Dict[str, Any]]",
        truth_format: "Union[str, List[str]]",
        truth_format_configuration: "Optional[Dict[str, Any]]",
        measures: "List[str]",
    ):
        self._run_format = run_format
        self._run_format_configuration = run_format_configuration
        self._truth_format = truth_format
        self._truth_format_configuration = truth_format_configuration
        self._measures = measures

    def evaluate(self, run: Path, truths: Path) -> "dict[str, Any]":
        self.is_valid(run, self._run_format, self._run_format_configuration, True)
        self.is_valid(truths, self._truth_format, self._truth_format_configuration)

        run_data = lines_if_valid(run, self._run_format, self._run_format_configuration)
        truth_data = lines_if_valid(truths, self._truth_format, self._truth_format_configuration)

        return self._eval(run_data, truth_data)

    def is_valid(
        self,
        directory: Path,
        format: "Union[str, List[str]]",
        configuration: "Optional[Dict[str, Any]]" = None,
        log: bool = False,
    ):
        ret = check_format(directory, format, configuration)
        if log:
            log_message(ret[1], ret[0])
        if ret[0] != _fmt.OK:

            raise ValueError(ret[1])

    def throw_if_conf_invalid(self, config: dict) -> None:
        raise ValueError("This is not implemented")

    def _eval(self, run_data: List[dict], truth_data: List[dict]) -> dict:
        raise ValueError("This is not implemented")


class TextGenerationEvaluator(TiraBaseEvaluator):
    def throw_if_conf_invalid(self, config: dict) -> None:
        if not self._run_format:
            raise ValueError(f"Configuration error. I need a run format. Got: {self._run_format}")

        run_format_checker = check_format_configuration_if_valid(self._run_format, self._run_format_configuration)

        if not run_format_checker.has_id_and_value_field():
            raise ValueError(f"The run format needs an field {CONF_ID_FIELD} and an field {CONF_VALUE_FIELD}")

        self.run_label_column = run_format_checker.value_field
        self.run_id_column = run_format_checker.id_field

        self.truth_label_column = None
        self.truth_id_column = None

        if self._truth_format:
            truth_format_checker = check_format_configuration_if_valid(
                self._truth_format, self._truth_format_configuration
            )
            if truth_format_checker.id_field is not None:
                self.truth_id_column = truth_format_checker.id_field

    def evaluate(self, run: Path, truths: Path) -> "dict[str, Any]":
        self.is_valid(run, self._run_format, self._run_format_configuration, True)

        run_data = lines_if_valid(run, self._run_format, self._run_format_configuration)
        truth_data = None

        if self.truth_id_column:
            self.is_valid(truths, self._truth_format, self._truth_format_configuration)
            truth_data = lines_if_valid(truths, self._truth_format, self._truth_format_configuration)

        return self._eval(run_data, truth_data)

    def _eval(self, run_data: List[dict], truth_data: List[dict]) -> dict:
        id_to_length = {}
        for i in run_data:
            if i[self.run_id_column] in id_to_length:
                raise ValueError("The run has duplicated IDs. The Id '{i[self.run_id_column]}' is duplicated")
            id_to_length[i[self.run_id_column]] = len(i[self.run_label_column].split())

        if truth_data and self.truth_id_column:
            for i in truth_data:
                if i[self.truth_id_column] not in id_to_length:
                    raise ValueError(f"No text was generated for the {self.run_id_column}={i[self.truth_id_column]}.")

        ret = {
            "Words (Min)": min(id_to_length.values()),
            "Words (Avg)": mean(id_to_length.values()),
            "Words (Max)": max(id_to_length.values()),
        }

        return {k: v for k, v in ret.items() if k in self._measures}


class RunFileEvaluator(TiraBaseEvaluator):
    def throw_if_conf_invalid(self, config: dict) -> None:

        if ("run.txt" != self._run_format and "run.txt" not in self._run_format) and (
            "LongEvalLags" != self._run_format and "LongEvalLags" not in self._run_format
        ):
            raise ValueError("I can only use the RunFileEvaluator for run.txt/LongEvalLags format")

        self.lags = [""]
        if "LongEvalLags" == self._run_format or "LongEvalLags" in self._run_format:
            self.lags = config["lags"]

        self._run_format = "run.txt"
        if self._truth_format and "qrels.txt" != self._truth_format and "qrels.txt" not in self._truth_format:
            self._truth_format = "qrels.txt"

    def evaluate(self, run_orig: Path, truths: Path) -> "dict[str, Any]":
        ret = {}

        for lag in self.lags:
            run = run_orig
            if lag:
                run = Path(run_orig) / lag
            self.is_valid(Path(run), self._run_format, True)

            expected_queries = None
            if self._truth_format is not None:

                self.is_valid(truths, self._truth_format)
                expected_queries = lines_if_valid(truths, self._truth_format)
                expected_queries = set([i["qid"] for i in expected_queries])

            run_data = lines_if_valid(run, self._run_format)
            counts = defaultdict(set)

            for i in run_data:
                if expected_queries and i["qid"] not in expected_queries:
                    continue
                counts[i["qid"]].add(i["docno"])

            lengths = [len(i) for i in counts.values()]
            num_queries = len(counts.keys())
            prefix = "" if not lag else f"{lag} "

            ret.update(
                {
                    f"{prefix}Docs Per Query (Avg)": sum(lengths) / num_queries,
                    f"{prefix}Docs Per Query (Min)": min(lengths),
                    f"{prefix}Docs Per Query (Max)": max(lengths),
                    f"{prefix}NumQueries": num_queries,
                }
            )

        return {k: ret[k] for k in ret if any([i in k for i in self._measures])}


class WowsEvalEvaluator(TiraBaseEvaluator):
    def throw_if_conf_invalid(self, config: dict) -> None:
        pass

    def normalize_data(self, df: Any) -> Any:
        import pandas as pd

        if isinstance(df, pd.DataFrame):
            return self.normalize_data([i.to_dict() for _, i in df.iterrows()])
        else:
            ret = []
            for i in df.copy():
                i = i.copy()
                for field_to_delete in ["unknown"]:
                    if field_to_delete in i:
                        del i[field_to_delete]
                ret.append(i)
            return ret

    def __pointwise_rankings(self, id_to_query_doc, predictions):
        truths_rankings = {}
        predictions_rankings = {}

        for k, v in id_to_query_doc.items():
            if v["query_id"] not in truths_rankings:
                truths_rankings[v["query_id"]] = []
                predictions_rankings[v["query_id"]] = []

            truths_rankings[v["query_id"]].append({"doc_id": v["doc_id"], "score": v["qrel"]})
            predictions_rankings[v["query_id"]].append(
                {"doc_id": v["doc_id"], "score": float(predictions[k]["probability_relevant"])}
            )

        return truths_rankings, predictions_rankings

    def __pairwise_rankings(self, id_to_query_doc: Any, predictions: Any):
        truths_rankings = {}
        predictions_rankings = {}

        for k, v in id_to_query_doc.items():
            if v["query_id"] not in truths_rankings:
                truths_rankings[v["query_id"]] = {}
                predictions_rankings[v["query_id"]] = {}

            if v["doc_id"] not in truths_rankings[v["query_id"]]:
                truths_rankings[v["query_id"]][v["doc_id"]] = v["qrel"]

            if v["doc_id"] not in predictions_rankings[v["query_id"]]:
                predictions_rankings[v["query_id"]][v["doc_id"]] = 0

            predictions_rankings[v["query_id"]][v["doc_id"]] += float(predictions[k]["probability_relevant"])

        ret_truths_ranking = {}
        ret_predictions_rankings = {}

        for qid, docids in truths_rankings.items():
            ret_truths_ranking[qid] = [{"doc_id": i, "score": truths_rankings[qid][i]} for i in docids]

        for qid, docids in predictions_rankings.items():
            ret_predictions_rankings[qid] = [{"doc_id": i, "score": predictions_rankings[qid][i]} for i in docids]

        return ret_truths_ranking, ret_predictions_rankings

    def __sorted(self, ret):
        import pandas as pd

        ret = pd.DataFrame(ret)
        ret = ret.sort_values(["score", "doc_id"], ascending=[False, True])
        return [(i["score"], i["doc_id"]) for _, i in ret.iterrows()]

    def _eval(self, run_data: List[dict], truth_data: List[dict]) -> dict:
        id_to_query_doc = {}
        pairwise = False
        predictions = self.normalize_data(run_data)
        truths = self.normalize_data(truth_data)
        from trectools import misc

        for i in truths:
            id_to_query_doc[i["id"]] = {
                "query_id": i["query_id"],
                "doc_id": i["unknown_doc_id"],
                "qrel": int(i["qrel_unknown_doc"]),
            }
            if "relevant_doc_id" in i:
                pairwise = True
                id_to_query_doc[i["id"]]["relevant_doc_id"] = i["relevant_doc_id"]

        predictions = {i["id"]: i for i in predictions}

        if predictions.keys() != id_to_query_doc.keys():
            raise ValueError("fooo")

        if pairwise:
            truths_rankings, predictions_rankings = self.__pairwise_rankings(id_to_query_doc, predictions)
        else:
            truths_rankings, predictions_rankings = self.__pointwise_rankings(id_to_query_doc, predictions)

        tau_ap = []
        kendall = []
        spearman = []
        pearson = []

        for query_id in truths_rankings.keys():
            truth_ranking = self.__sorted(truths_rankings[query_id])
            predicted_ranking = self.__sorted(predictions_rankings[query_id])

            tau_ap.append(misc.get_correlation(truth_ranking, predicted_ranking, correlation="tauap")[0])
            kendall.append(misc.get_correlation(truth_ranking, predicted_ranking, correlation="kendall")[0])
            spearman.append(misc.get_correlation(truth_ranking, predicted_ranking, correlation="spearman")[0])
            pearson.append(misc.get_correlation(truth_ranking, predicted_ranking, correlation="pearson")[0])

        ret = {
            "wows_tau_ap": mean(tau_ap),
            "wows_kendall": mean(kendall),
            "wows_spearman": mean(spearman),
            "wows_pearson": mean(pearson),
        }
        return {i: ret[i] for i in self._measures}


class HuggingFaceEvaluator(TiraBaseEvaluator):
    def throw_if_conf_invalid(self, config: dict) -> None:
        if not self._run_format or not self._truth_format:
            raise ValueError(
                f"Configuration error. I need a run and truth format. Got: run_format={self._run_format} and truth_format={self._truth_format}."
            )

        run_format_checker = check_format_configuration_if_valid(self._run_format, self._run_format_configuration)

        if not run_format_checker.has_id_and_value_field():
            raise ValueError(f"The run format needs an field {CONF_ID_FIELD} and an field {CONF_VALUE_FIELD}")

        self.run_label_column = run_format_checker.value_field
        self.run_id_column = run_format_checker.id_field

        truth_format_checker = check_format_configuration_if_valid(self._truth_format, self._truth_format_configuration)

        if not truth_format_checker.has_id_and_value_field():
            raise ValueError(f"The truth format needs an field {CONF_ID_FIELD} and an field {CONF_VALUE_FIELD}")

        self.truth_label_column = truth_format_checker.value_field
        self.truth_id_column = truth_format_checker.id_field
        if "additional_args" in config and config["additional_args"]:
            self.additional_args = config["additional_args"]
        else:
            self.additional_args = {}

        import evaluate
        from evaluate.utils.file_utils import DownloadConfig

        if os.environ.get("OFFLINE", False):
            evaluate.config.HF_EVALUATE_OFFLINE = True

    def _eval(self, run_data: List[dict], truth_data: List[dict]) -> dict:
        run_data = [{"id": i[self.run_id_column], "prediction": i[self.run_label_column]} for i in run_data]
        truth_data = [{"id": i[self.truth_id_column], "truth": i[self.truth_label_column]} for i in truth_data]

        import evaluate
        import pandas as pd
        from evaluate.utils.file_utils import DownloadConfig

        download_config = None
        if os.environ.get("OFFLINE", False):
            download_config = DownloadConfig(local_files_only=True)

        df = pd.merge(pd.DataFrame(run_data), pd.DataFrame(truth_data), left_index=True, right_index=True)

        run_data = df["prediction"].tolist()
        truth_data = df["truth"].tolist()

        ret = {}
        for m in self._measures:
            ret[m] = evaluate.load(m, download_config=download_config).compute(
                predictions=run_data, references=truth_data, **self.additional_args
            )[m]
        return ret


class TrecToolsEvaluator(TiraBaseEvaluator):
    def throw_if_conf_invalid(self, config: dict) -> None:

        if ("LongEvalLags" == self._run_format or "LongEvalLags" in self._run_format) and (
            "LongEvalLags" == self._truth_format or "LongEvalLags" in self._truth_format
        ):
            self.groups_to_evaluate = self._run_format_configuration["lags"]
            self._run_format = "run.txt"
            self._truth_format = "qrels.txt"
        else:
            if "run.txt" != self._run_format and "run.txt" not in self._run_format:
                raise ValueError("I can only use trectools for run.txt format")

            if "qrels.txt" != self._run_format and "qrels.txt" not in self._truth_format:
                raise ValueError("I can only use trectools for run.txt format")

            self._run_format = "run.txt"
            self._truth_format = "qrels.txt"
            self.groups_to_evaluate = None

        # check that dependencies are available
        import pandas as pd
        from trectools import TrecEval, TrecQrel, TrecRun

    def evaluate(self, run: Path, truths: Path) -> "dict[str, Any]":
        if not self.groups_to_evaluate:
            return super().evaluate(run, truths)
        else:
            ret = {}
            for group in self.groups_to_evaluate:
                ret[group] = super().evaluate(run / group, truths / group)
            return ret

    def _eval(self, run_data: List[Any], truth_data: List[Any]) -> dict:
        import pandas as pd
        from trectools import TrecEval, TrecQrel, TrecRun

        run = TrecRun()
        run.run_data = pd.DataFrame(run_data)
        run.run_data["query"] = run.run_data["qid"]
        run.run_data["docid"] = run.run_data["docno"]

        qrels = TrecQrel()
        qrels.qrels_data = pd.DataFrame(truth_data)
        qrels.qrels_data["query"] = qrels.qrels_data["qid"]
        qrels.qrels_data["docid"] = qrels.qrels_data["docno"]

        te = TrecEval(run, qrels)

        ret = {}
        if "nDCG@10" in self._measures:
            ret["nDCG@10"] = te.get_ndcg(depth=10)

        if "P@10" in self._measures:
            ret["P@10"] = te.get_precision(depth=10)

        if "RR" in self._measures:
            ret["RR"] = te.get_reciprocal_rank()

        return {k: ret[k] for k in self._measures}


EVALUATORS: dict[str, TiraBaseEvaluator] = {
    "TrecTools": TrecToolsEvaluator,
    "RunFileEvaluator": RunFileEvaluator,
    "HuggingFaceEvaluator": HuggingFaceEvaluator,
    "WowsEvalEvaluator": WowsEvalEvaluator,
    "TextGenerationEvaluator": TextGenerationEvaluator,
}

MEASURE_TO_EVALUATORS: dict[str, str] = {
    "nDCG@10": "TrecTools",
    "RR": "TrecTools",
    "P@10": "TrecTools",
    "Docs Per Query (Avg)": "RunFileEvaluator",
    "Docs Per Query (Min)": "RunFileEvaluator",
    "Docs Per Query (Max)": "RunFileEvaluator",
    "NumQueries": "RunFileEvaluator",
    "accuracy": "HuggingFaceEvaluator",
    "recall": "HuggingFaceEvaluator",
    "precision": "HuggingFaceEvaluator",
    "f1": "HuggingFaceEvaluator",
    "wows_tau_ap": "WowsEvalEvaluator",
    "wows_kendall": "WowsEvalEvaluator",
    "wows_pearson": "WowsEvalEvaluator",
    "wows_spearman": "WowsEvalEvaluator",
    "Words (Min)": "TextGenerationEvaluator",
    "Words (Avg)": "TextGenerationEvaluator",
    "Words (Max)": "TextGenerationEvaluator",
}


def load_evaluator_config(config: "Union[dict, str]", client: "Optional[TiraClient]" = None) -> dict:
    if isinstance(config, str):
        if client is None:
            client = Client()
        dataset_config = client.get_dataset(config)
        if "trusted_evaluator" not in dataset_config or not dataset_config["trusted_evaluator"]:
            raise ValueError(f'No trusted evaluation is configured for the dataset "{config}".')

        return load_evaluator_config(dataset_config["trusted_evaluator"])

    config = config.copy()

    if "measures" not in config or not config["measures"]:
        if "trusted_eval" in config and config["trusted_eval"] and "measures" in config["trusted_eval"]:
            config["measures"] = config["trusted_eval"]["measures"]
        else:
            raise ValueError("Configuration of the evaluator is invalid: No measures are specified.")

    if "run_format" not in config:
        if "format" in config:
            config["run_format"] = config["format"]
        else:
            raise ValueError("Configuration of the evaluator is invalid: No run_format is specified.")

    if "truth_format" not in config:
        raise ValueError("Configuration of the evaluator is invalid: No truth_format is specified.")

    return config


def unsandboxed_evaluation_is_allowed(config: "Union[dict, str]", client: "Optional[TiraClient]" = None) -> bool:
    try:
        config = load_evaluator_config(config, client)

        return config and "measures" in config and config["measures"]
    except:
        pass
    return False


def get_evaluators_if_valid(
    config: "Union[dict, str]", client: "Optional[TiraClient]" = None
) -> List[TiraBaseEvaluator]:
    config = load_evaluator_config(config, client)

    evaluator_to_measures = defaultdict(list)
    for measure in config["measures"]:
        evaluator = MEASURE_TO_EVALUATORS[measure]
        evaluator_to_measures[evaluator].append(measure)

    ret = []
    for evaluator, measures in evaluator_to_measures.items():
        format_config = config.get("run_format_configuration")
        if not format_config and "format_configuration" in config:
            format_config = config.get("format_configuration")

        impl = EVALUATORS[evaluator](
            config["run_format"],
            format_config,
            config["truth_format"],
            config.get("truth_format_configuration"),
            measures,
        )
        impl.throw_if_conf_invalid(config)
        ret.append(impl)

    return ret


def evaluate(
    run: Path,
    truths: Path,
    config: "Union[dict, str]",
    output_dir: "Optional[Path]" = None,
    client: "Optional[TiraClient]" = None,
    monitored: "Optional[bool]" = False,
) -> dict:
    config = load_evaluator_config(config, client)
    evaluators = get_evaluators_if_valid(config, client)

    if monitored:
        from tira.io_utils import MonitoredExecution

        return MonitoredExecution().run(lambda i: evaluate(run, truths, config, i))

    ret = {}
    for evaluator in evaluators:
        evaluation = evaluator.evaluate(run, truths)
        ret.update(evaluation)

    if output_dir:
        prototext = to_prototext([ret])
        (output_dir / "evaluation.prototext").write_text(prototext)

    return ret



import json
import unittest

from . import _ERROR, _OK, EMPTY_OUTPUT, TEXT_ALIGNMENT_FEATURES_VALID


def check_format(dir):
    from tira.check_format import check_format as cf

    return cf(dir, "text-alignment-features")


def lines_if_valid(dir):
    from tira.check_format import lines_if_valid as lf

    return lf(dir, "text-alignment-features")


class TestCheckTextAlignmentFeaturesFormats(unittest.TestCase):
    def test_invalid_validator_on_empty_output(self):
        expected = [_ERROR, "The text-alignment-feature directory contains only 0 instances, this is likely an error."]
        actual = check_format(EMPTY_OUTPUT)
        self.assertEqual(expected, actual)

    def test_valid_directory(self):
        expected = [_OK, "The directory has the correct format."]
        actual = check_format(TEXT_ALIGNMENT_FEATURES_VALID)
        self.assertEqual(expected, actual)

    def test_parsed_lines(self):
        lines = lines_if_valid(TEXT_ALIGNMENT_FEATURES_VALID)
        expected = sorted(
            [
                "suspicious-document00028.txt",
                "suspicious-document00028.txt",
                "suspicious-document00172.txt",
                "suspicious-document00172.txt",
            ]
        )
        self.assertEqual(sorted([i["this_reference"] for i in lines]), expected)



import unittest

from tira.io_utils import all_environment_variables_for_github_action_or_fail


class TestExportSubmissionEnvironment(unittest.TestCase):
    def test_empty_parameters_fails(self):
        with self.assertRaises(ValueError):
            all_environment_variables_for_github_action_or_fail([])

    def test_all_required_parameters_are_there_docker_image(self):
        actual = all_environment_variables_for_github_action_or_fail(
            [
                "GITHUB_SHA=sha",
                "TIRA_VM_ID=vm",
                "TIRA_TASK_ID=task",
                "TIRA_DOCKER_REGISTRY_TOKEN=docker-token",
                "TIRA_DOCKER_REGISTRY_USER=docker-user",
                "TIRA_CLIENT_TOKEN=tira-token",
                "TIRA_CLIENT_USER=tira-user",
                "TIRA_CODE_REPOSITORY_ID=repo-id",
                "TIRA_DOCKER_FILE=Dockerfile",
                "TIRA_DOCKER_PATH=dockerpath",
                "TIRA_COMMAND=cmd",
            ]
        )

        self.assertTrue("GITHUB_SHA=sha" in actual)
        self.assertTrue(
            "IMAGE_TAG=registry.webis.de/code-research/tira/tira-user-vm/submission-dockerpath:sha" in actual
        )

    def test_all_required_parameters_are_there_jupyter_notebook(self):
        actual = all_environment_variables_for_github_action_or_fail(
            [
                "GITHUB_SHA=sha",
                "TIRA_VM_ID=vm",
                "TIRA_TASK_ID=task",
                "TIRA_DOCKER_REGISTRY_TOKEN=docker-token",
                "TIRA_DOCKER_REGISTRY_USER=docker-user",
                "TIRA_CLIENT_TOKEN=tira-token",
                "TIRA_CLIENT_USER=tira-user",
                "TIRA_CODE_REPOSITORY_ID=repo-id",
                "TIRA_DOCKER_FILE=Dockerfile",
                "TIRA_DOCKER_PATH=dockerpath",
                "TIRA_COMMAND=cmd",
                "TIRA_JUPYTER_NOTEBOOK=123",
            ]
        )

        self.assertTrue("GITHUB_SHA=sha" in actual)
        self.assertTrue("IMAGE_TAG=registry.webis.de/code-research/tira/tira-user-vm/submission:sha" in actual)



import unittest

import pandas as pd

from tira.third_party_integrations import persist_and_normalize_run

# .. todo:: there are no assertions in this file


class TestPersistingRunFiles(unittest.TestCase):

    def test_variant_01(self):
        run = pd.DataFrame(
            [
                {"qid": "1", "score": "1", "docno": "d1"},
                {"qid": "1", "score": "0", "docno": "d2"},
                {"qid": "1", "score": "2", "docno": "d3"},
            ]
        )
        persist_and_normalize_run(run, "system_name", default_output="/tmp/variant_01.run.txt")

    def test_variant_02(self):
        run = pd.DataFrame(
            [
                {"qid": "1", "score": "1", "docno": "d1"},
                {"qid": "1", "score": "0", "docno": "d2"},
                {"qid": "1", "score": "2", "docno": "d3"},
            ]
        )
        persist_and_normalize_run(run, "system_name", output_file="/tmp/variant_02.run.txt")

    def test_variant_03(self):
        run = pd.DataFrame(
            [
                {"qid": "1", "score": "1", "docno": "d1"},
                {"qid": "1", "score": "0", "docno": "d2"},
                {"qid": "1", "score": "2", "docno": "d3"},
            ]
        )
        persist_and_normalize_run(run, "system_name")



import unittest

from tira.local_execution_integration import LocalExecutionIntegration

# .. todo:: there are no assertions in this file


class TestExporSubmissionFromJupyterNotebook(unittest.TestCase):
    def test_non_existing_notebook_1(self):
        notebook = "does-not-exist"

        actual = LocalExecutionIntegration().export_submission_from_jupyter_notebook(notebook)

        assert actual is None

    def test_non_existing_notebook_2(self):
        notebook = "does-not-exist/1/does-not-exist.ipynb"

        actual = LocalExecutionIntegration().export_submission_from_jupyter_notebook(notebook)

        assert actual is None

    def test_notebook_submission_without_previous_stages(self):
        notebook = "tests/resources/pyterrier-notebook-without-previous-stages.ipynb"

        expected = (
            "TIRA_COMMAND=/workspace/run-pyterrier-notebook.py "
            "--input ${TIRA_INPUT_DIRECTORY} "
            "--output ${TIRA_OUTPUT_DIRECTORY} "
            "--notebook /workspace/pyterrier-notebook-without-previous-stages.ipynb"
        )
        actual = LocalExecutionIntegration().export_submission_from_jupyter_notebook(notebook)

        assert expected == actual



#!/usr/bin/env python3
import logging
import os
import subprocess
import sys
from argparse import ArgumentParser, ArgumentTypeError

from tira.inference_server import run_inference_server


def limited_int_arg(lower_bound: int, upper_bound: int):
    def ret(arg: str) -> int:
        i = int(arg)
        if not (lower_bound <= i <= upper_bound):
            raise ArgumentTypeError(f"{i} is outside the permissible range from {lower_bound} to {upper_bound}")
        return i

    return ret


def parse_args():
    parser = ArgumentParser(prog="tira-run-inference-server", description="")

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--script", type=str, help="The python script containing the model and the predict function.", required=False
    )
    group.add_argument(
        "--notebook", type=str, help="The notebook containing the model and the predict function.", required=False
    )

    parser.add_argument("--port", type=limited_int_arg(5000, 65536), help="Port number of local server.", required=True)
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Sets the output verbosity. Default is INFO. Can be repeated for more verbose output.",
    )

    return parser.parse_args()


def main():
    args = parse_args()

    if args.notebook is not None:
        notebook_name = args.notebook
        # convert notebook to python script
        command_convert = f"jupyter nbconvert --to script {notebook_name}"
        try:
            return_code = subprocess.check_call(
                command_convert, shell=True, stdout=sys.stdout, stderr=subprocess.STDOUT
            )
        except subprocess.CalledProcessError:
            return_code = 2
        if return_code != 0:
            print(f"Exception while converting notebook '{notebook_name}'. Stopping server.")
            sys.exit(return_code)
        module_name = notebook_name
    else:
        module_name = args.script

    absolute_path = os.path.splitext(os.path.abspath(module_name))[0] + ".py"
    module_name = os.path.splitext(module_name)[0].replace("/", ".")

    # flag execution for running as inference server (see third_party_integrations.is_running_as_inference_server)
    os.environ["TIRA_INFERENCE_SERVER"] = "True"

    loglevels = [logging.INFO, logging.DEBUG]
    loglevel = loglevels[min(args.verbose, len(loglevels) - 1)]
    run_inference_server(
        base_module=module_name, absolute_path=absolute_path, internal_port=args.port, loglevel=loglevel
    )



import json
import zipfile
from datetime import datetime as dt
from pathlib import Path

from tira.tira_client import TiraClient


class ProfilingIntegration:
    """
    Access the profiling of runs executed in TIRA, e.g., CPU and memory usage, but als GPU utilization if available.
    """

    def __init__(self, tira_client: TiraClient):
        """Instantiate the ProfilingIntegration that uses the passed tira_client to acccess runs and parse their
        profiling metadata.

        Args:
            tira_client (TiraClient): the tira client to access the runs and their profiling metadata.
        """
        self.tira_client = tira_client

    def from_submission(
        self, approach: str, dataset: str, return_pd: bool = False, allow_without_evaluation: bool = False
    ):
        """Return the profiling of the run identified by the approach on the dataset, i.e.,  CPU and memory usage, but
        als GPU utilization if available. Will throw an exception if no profiling data is available (e.g., if profiling
        was not configured for the task).

        Entries look like [{"timestamp": 0.0, "key": "ps_cpu", "value": 0.3}, ...]. The timestamp is the time in
        seconds since the start of the run, the key is the name of the metric, and the value is the value of the
        metric. The following metrics can be available (depending on the run and the system configuration):

        - elapsed_time: elapsed time in seconds since the start of the run until completion of the process.
        - ps_cpu: CPU usage in percent, produced by the `ps` command.
        - ps_rss: RSS Memory usage, produced by the `ps` command.
        - ps_vsz: VSZ Memory usage, produced by the `ps` command.
        - gpu_memory_used: Memory usage of the GPU in MiB, produced by the `nvidia-smi` command.
        - gpu_utilization: Utilization of the GPU in percent, produced by the `nvidia-smi` command.

        Args:
            approach (str): The identifier of the approach, e.g., "<team>/<task>/<approach>".
            dataset (str): The dataset identifier, e.g., "reneuir-2024/dl-top-1000-docs-20240701-training".
            return_pd (str, optional): Return as pandas DataFrame instead of as list of dictionaries. Defaults to False.
            allow_without_evaluation (bool, optional): allow to retrieve runs without evaluation. Defaults to False.
        """

        try:
            run_output_dir = self.tira_client.get_run_output(approach, dataset, allow_without_evaluation)
            run_output_dir = Path(run_output_dir).parent
        except Exception as e:
            raise Exception(
                f"No profiling data available for approach '{approach}' on dataset '{dataset}'. Could not load run", e
            )

        return self.from_local_run_output(run_output_dir, return_pd)

    def raw_telemetry(self, approach: str, dataset: str, resource: str, allow_without_evaluation: bool = False) -> str:
        """Return the raw telemetry "resource" of the run identified by the approach on the dataset. The passed
        resource specifies which telemetry to return, i.e.,

        - cpuinfo: The content of '/proc/cpuinfo' of the host that executed the run.
        - meminfo: The content of '/proc/meminfo' of the host that executed the run.
        - nvidia-smi.out: The content of the 'nvidia-smi' command of the host that executed the run, executed once
            before the software was started.
        - nvidia-smi.log: Periodic telemetry of nvidia-smi monitored while the software was executed in the sandbox.
        - ps.log: Periodic telemetry of 'ps' monitored while the software was executed in the sandbox.

        Args:
            approach (str): The identifier of the approach, e.g., "<team>/<task>/<approach>".
            dataset (str): The dataset identifier, e.g., "reneuir-2024/dl-top-1000-docs-20240701-training".
            resource (str): the telemetry to return.
            allow_without_evaluation (bool, optional): allow to retrieve runs without evaluation. Defaults to False.
        """
        try:
            run_output_dir = self.tira_client.get_run_output(approach, dataset, allow_without_evaluation)
        except Exception as e:
            raise Exception(
                f"No profiling data available for approach '{approach}' on dataset '{dataset}'. Could not load run", e
            )

        return self._read_file_from_profiling_zip(Path(run_output_dir).parent / "profiling.zip", resource)

    def _read_file_from_profiling_zip(self, profiling_zip: Path, file: str):
        with zipfile.ZipFile(profiling_zip, "r") as archive:
            return archive.read(file).decode("utf-8")

    def from_local_run_output(self, run_output_dir: Path, return_pd: bool = False):
        """Return the profiling of the run within the run output dir, i.e.,  CPU and memory usage, but als GPU
        utilization if available. Will throw an exception if no profiling data is available (e.g., if profiling was not
        configured for the task).

        Entries look like [{"timestamp": 0.0, "key": "ps_cpu", "value": 0.3}, ...]. The timestamp is the time in
        seconds since the start of the run, the key is the name of the metric, and the value is the value of the
        metric. The following metrics can be available (depending on the run and the system configuration):

        - elapsed_time: elapsed time in seconds since the start of the run until completion of the process.
        - ps_cpu: CPU usage in percent, produced by the `ps` command.
        - ps_rss: RSS Memory usage, produced by the `ps` command.
        - ps_vsz: VSZ Memory usage, produced by the `ps` command.
        - gpu_memory_used: Memory usage of the GPU in MiB, produced by the `nvidia-smi` command.
        - gpu_utilization: Utilization of the GPU in percent, produced by the `nvidia-smi` command.

        Args:
            run_output_dir (Path): The path to the output dir.
            return_pd (str, optional): Return as pandas DataFrame instead of as list of dictionaries. Defaults to False.
        """
        profiling_file = run_output_dir / "parsed_profiling.jsonl"
        start_time = run_output_dir / "start"
        end_time = run_output_dir / "end"
        profiling_zip = run_output_dir / "profiling.zip"

        if not profiling_file.exists() or (
            (not start_time.exists() or not end_time.exists()) and not profiling_zip.exists()
        ):
            raise Exception(f"No profiling data available for run {run_output_dir}.")

        try:
            start_time = open(start_time).read()
            end_time = open(end_time).read()
        except Exception:
            start_time = self._read_file_from_profiling_zip(profiling_zip, "start")
            end_time = self._read_file_from_profiling_zip(profiling_zip, "end")

        start_time = " ".join(start_time.split()[:4])
        end_time = " ".join(end_time.split()[:4])
        start_time = dt.strptime(start_time, "%a %b %d %H:%M:%S").timestamp()
        end_time = dt.strptime(end_time, "%a %b %d %H:%M:%S").timestamp()
        ret = []
        with open(profiling_file, "r") as f:
            for line in f:
                ret.append(json.loads(line))

        ret.append({"timestamp": end_time - start_time, "key": "elapsed_time", "value": end_time - start_time})

        if return_pd:
            import pandas as pd

            return pd.DataFrame(ret)
        else:
            return ret



import unittest

from tira.check_format import _fmt, check_format
from tira.rest_api_client import Client

ALL_DATASETS = {}
tira = Client()
TASKS = ["wows-eval", "advertisement-in-retrieval-augmented-generation-2025"]
for task in TASKS:
    ALL_DATASETS.update(tira.datasets(task))

import tempfile
from pathlib import Path

from parameterized import parameterized


def datasets_with_format(dataset_type):
    ret = {}
    for k, v in ALL_DATASETS.items():
        if f"{dataset_type}_format" in v:
            ret[k] = {
                f"{dataset_type}_format": v[f"{dataset_type}_format"],
                f"{dataset_type}_format_configuration": v.get(f"{dataset_type}_format_configuration"),
            }
    return ret


MINIMAL_WOWS_POINTWISE = {
    "run": """{"id": "32d23068-7440-4891-9958-42325f98a604", "probability_relevant": 0.5}
{"id": "cde83146-ac3e-4bc5-a959-f2006ac7b8de", "probability_relevant": 0.5}
{"id": "cb7b20d0-def6-46c4-ae44-a78f00b47735", "probability_relevant": 0.5}
{"id": "4a68c86f-64ea-4293-bda5-7a0130c13864", "probability_relevant": 0.5}
{"id": "3e550de5-a104-44ae-bc3f-7ab556cc1018", "probability_relevant": 0.5}""",
    "truth": """{"id": "32d23068-7440-4891-9958-42325f98a604", "query": "who sings monk theme song", "query_id": 1051399, "unknown_doc_id": "4426187", "qrel_unknown_doc": 3}
{"id": "cde83146-ac3e-4bc5-a959-f2006ac7b8de", "query": "who sings monk theme song", "query_id": 1051399, "unknown_doc_id": "4642930", "qrel_unknown_doc": 0}
{"id": "cb7b20d0-def6-46c4-ae44-a78f00b47735", "query": "who sings monk theme song", "query_id": 1051399, "unknown_doc_id": "3108511", "qrel_unknown_doc": 1}
{"id": "4a68c86f-64ea-4293-bda5-7a0130c13864", "query": "who sings monk theme song", "query_id": 1051399, "unknown_doc_id": "5040048", "qrel_unknown_doc": 1}
{"id": "3e550de5-a104-44ae-bc3f-7ab556cc1018", "query": "who sings monk theme song", "query_id": 1051399, "unknown_doc_id": "3376628", "qrel_unknown_doc": 0}""",
}

MINIMAL_WOWS_PAIRWISE = {
    "run": """{"id": "3d080873-98a1-4388-af86-fe2c8b47ebca", "probability_relevant": 0.5}
{"id": "468a9e92-467f-47c9-810b-fe6fa9dca634", "probability_relevant": 0.5}
{"id": "846a69d0-0c0e-4d86-baf2-c3e8d31fdc86", "probability_relevant": 0.5}
{"id": "83c0e22c-b00f-4570-b1cb-027199c673d4", "probability_relevant": 0.5}
{"id": "a88a0a31-4795-4c59-830b-848de52a7fd6", "probability_relevant": 0.5}""",
    "truth": """{"id": "3d080873-98a1-4388-af86-fe2c8b47ebca", "query": "who sings monk theme song", "query_id": 1051399, "unknown_doc_id": "2378828", "qrel_unknown_doc": 2, "relevant_doc_id": "69813"}
{"id": "468a9e92-467f-47c9-810b-fe6fa9dca634", "query": "who sings monk theme song", "query_id": 1051399, "relevant_doc_id": "69813", "unknown_doc_id": "3376628", "qrel_unknown_doc": 0}
{"id": "846a69d0-0c0e-4d86-baf2-c3e8d31fdc86", "query": "who sings monk theme song", "query_id": 1051399, "relevant_doc_id": "69813", "unknown_doc_id": "5040048", "qrel_unknown_doc": 1}
{"id": "83c0e22c-b00f-4570-b1cb-027199c673d4", "query": "who sings monk theme song", "query_id": 1051399,  "relevant_doc_id": "69813", "unknown_doc_id": "4642930", "qrel_unknown_doc": 0}
{"id": "a88a0a31-4795-4c59-830b-848de52a7fd6", "query": "who sings monk theme song", "query_id": 1051399, "relevant_doc_id": "69813", "unknown_doc_id": "4426187", "qrel_unknown_doc": 3, "relevant_doc_id": "69813"}""",
}

MINIMAL_AD_CLASSIFICATION = {
    "run": """{"id": "xyz1", "label": 1, "tag": 1}
{"id": "xyz2", "label": 0, "tag": 1}
{"id": "xyz3", "label": 0, "tag": 1}""",
    "truth": """{"id": "xyz1", "label": 1, "tag": 1}
{"id": "xyz2", "label": 0, "tag": 1}
{"id": "xyz3", "label": 1, "tag": 1}""",
}

MINIMAL_AD_GENERATION = {
    "run": """{"id": "xyz1", "response": "1", "tag": "a", "topic": "a", "references": [], "advertisement": []}
{"id": "xyz2", "response": "1", "tag": "a", "topic": "a", "references": [], "advertisement": []}
{"id": "xyz3", "response": "1", "tag": "a", "topic": "a", "references": [], "advertisement": []}""",
    "truth": """{"id": "xyz1", "response": "1"}
{"id": "xyz2", "response": "1"}
{"id": "xyz3", "response": "1"}""",
}

DATASET_TO_MINIMAL_EXAMPLE = {
    "pairwise-20250309-test": MINIMAL_WOWS_PAIRWISE,
    "pairwise-smoke-test-20250210-training": MINIMAL_WOWS_PAIRWISE,
    "pointwise-20250309-test": MINIMAL_WOWS_POINTWISE,
    "pointwise-smoke-test-20250128-training": MINIMAL_WOWS_POINTWISE,
    "ads-in-rag-task-1-generation-spot-check-20250423_1-training": MINIMAL_AD_GENERATION,
    "ads-in-rag-task-1-generation-test-20250428-test": MINIMAL_AD_GENERATION,
    "ads-in-rag-task-1-generation-training-20250423-training": MINIMAL_AD_GENERATION,
    "ads-in-rag-task-2-classification-spot-check-20250423-training": MINIMAL_AD_CLASSIFICATION,
    "ads-in-rag-task-2-classification-test-20250428-test": MINIMAL_AD_CLASSIFICATION,
    "ads-in-rag-task-2-classification-training-20250423-training": MINIMAL_AD_CLASSIFICATION,
}


class TestIntegration(unittest.TestCase):
    def test_datasets_with_run_format(self):
        expected = DATASET_TO_MINIMAL_EXAMPLE.keys()
        actual = datasets_with_format("run").keys()
        self.assertEqual(expected, actual)

    def test_datasets_with_truth_format(self):
        expected = DATASET_TO_MINIMAL_EXAMPLE.keys()
        actual = datasets_with_format("truth").keys()
        self.assertEqual(expected, actual)

    @parameterized.expand(datasets_with_format("truth").items())
    def test_truth_datasets_are_valid(self, k, v):
        TYPE = "truth"
        with tempfile.TemporaryDirectory() as d:
            (Path(d) / "labels.jsonl").write_text(DATASET_TO_MINIMAL_EXAMPLE[k][TYPE])
            actual = check_format(Path(d), v[f"{TYPE}_format"], v[f"{TYPE}_format_configuration"])
            self.assertEqual(_fmt.OK, actual[0], f"Problem in {k}: {actual[1]}")

    @parameterized.expand(datasets_with_format("run").items())
    def test_run_datasets_are_valid(self, k, v):
        TYPE = "run"
        with tempfile.TemporaryDirectory() as d:
            (Path(d) / "labels.jsonl").write_text(DATASET_TO_MINIMAL_EXAMPLE[k][TYPE])
            actual = check_format(Path(d), v[f"{TYPE}_format"], v[f"{TYPE}_format_configuration"])
            self.assertEqual(_fmt.OK, actual[0], f"Problem in {k}: {actual[1]}")



import unittest

import pandas as pd

from tira.local_client import Client


def query_segmentation(queries):
    queries = pd.DataFrame([{"qid": str(i)} for i in queries])
    tira = Client("tests/resources/")
    query_segmentation = tira.pt.transform_queries(
        "ir-benchmarks/webis-query-segmentation/wt-snp-baseline", dataset="d1"
    )

    ret = query_segmentation(queries)
    return {i["qid"]: i["segmentation"] for _, i in ret.iterrows()}


def query_segmentation_from_local_directory(queries):
    queries = pd.DataFrame([{"qid": str(i)} for i in queries])
    tira = Client()
    query_segmentation = tira.pt.transform_queries(
        "tests/resources/query-processing-outputs/query-segmentation/queries.jsonl", None
    )

    ret = query_segmentation(queries)
    return {i["qid"]: i["segmentation"] for _, i in ret.iterrows()}


class TestQueryProcessingLoader(unittest.TestCase):
    def test_for_loading_query_segmentations_for_multiple_queries(self):
        expected = {"350": ["health and", "computer terminals"], "353": ["antarctica", "exploration"]}

        actual = query_segmentation([350, 353])

        self.assertEqual(len(actual), 2)
        self.assertEqual(actual["350"], expected["350"])
        self.assertEqual(actual["353"], expected["353"])

    def test_for_loading_query_segmentations_for_multiple_queries_from_dir(self):
        expected = {"350": ["health and", "computer terminals"], "353": ["antarctica", "exploration"]}

        actual = query_segmentation_from_local_directory([350, 353])

        self.assertEqual(len(actual), 2)
        self.assertEqual(actual["350"], expected["350"])
        self.assertEqual(actual["353"], expected["353"])

    def test_for_loading_query_segmentations_for_single_query_from_dir(self):
        expected = {"351": ["falkland", "petroleum exploration"]}

        actual = query_segmentation_from_local_directory([351])

        self.assertEqual(len(actual), 1)
        self.assertEqual(actual["351"], expected["351"])

    def test_for_loading_query_segmentations_for_single_query(self):
        expected = {"351": ["falkland", "petroleum exploration"]}

        actual = query_segmentation([351])

        self.assertEqual(len(actual), 1)
        self.assertEqual(actual["351"], expected["351"])

    def test_pyterrier_can_be_loaded(self):
        from tira.third_party_integrations import ensure_pyterrier_is_loaded

        ensure_pyterrier_is_loaded()

    def test_rest_query_submission_with_rest_api_01(self):
        from tira.rest_api_client import Client

        tira = Client()

        q = tira.pt.transform_queries(
            "ir-benchmarks/ows/query-segmentation-wt-snp", "disks45-nocr-trec-robust-2004-20230209-training"
        )
        self.assertEqual(len(q(pd.DataFrame([{"qid": "303"}]))), 1)
        self.assertEqual(
            q(pd.DataFrame([{"qid": "303"}])).iloc[0].to_dict()["segmentation"], ["hubble telescope achievements"]
        )

    def test_rest_query_submission_with_rest_api_02(self):
        from tira.rest_api_client import Client

        tira = Client()

        q = tira.pt.transform_queries(
            "ir-benchmarks/ows/query-segmentation-hyb-a", "disks45-nocr-trec-robust-2004-20230209-training"
        )
        self.assertEqual(len(q(pd.DataFrame([{"qid": "303"}]))), 1)
        self.assertEqual(
            q(pd.DataFrame([{"qid": "303"}])).iloc[0].to_dict()["segmentation"], ["hubble telescope", "achievements"]
        )

    def test_rest_query_submission_with_rest_api_03(self):
        import pyterrier as pt

        from tira.rest_api_client import Client
        from tira.third_party_integrations import ensure_pyterrier_is_loaded

        ensure_pyterrier_is_loaded()
        tira = Client()

        q = tira.pt.transform_queries(
            "ir-benchmarks/fschlatt/query-health-classification", pt.get_dataset("irds:clueweb09/en/trec-web-2009")
        )
        self.assertEqual(len(q(pd.DataFrame([{"qid": "26"}]))), 1)
        self.assertEqual(q(pd.DataFrame([{"qid": "26"}])).iloc[0].to_dict()["mean_health_score"], 168.416)
        self.assertEqual(q(pd.DataFrame([{"qid": "26"}])).iloc[0].to_dict()["median_health_score"], 130.3137)

    def test_rest_query_submission_with_rest_api_and_custom_selection(self):
        import pyterrier as pt

        from tira.rest_api_client import Client
        from tira.third_party_integrations import ensure_pyterrier_is_loaded

        ensure_pyterrier_is_loaded()
        tira = Client()

        q = tira.pt.transform_queries(
            "ir-benchmarks/fschlatt/query-health-classification", pt.get_dataset("irds:clueweb09/en/trec-web-2009")
        )
        self.assertEqual(len(q(pd.DataFrame([{"qid": "26"}]))), 1)
        self.assertEqual(q(pd.DataFrame([{"qid": "26"}])).iloc[0].to_dict()["mean_health_score"], 168.416)
        self.assertEqual(q(pd.DataFrame([{"qid": "26"}])).iloc[0].to_dict()["median_health_score"], 130.3137)

    def test_rest_query_submission_with_rest_api_04(self):
        import pyterrier as pt

        from tira.rest_api_client import Client
        from tira.third_party_integrations import ensure_pyterrier_is_loaded

        ensure_pyterrier_is_loaded()
        tira = Client()

        q = tira.pt.transform_queries(
            "ir-benchmarks/qpptk/fixed-sealer", pt.get_dataset("irds:disks45/nocr/trec-robust-2004")
        )
        self.assertEqual(len(q(pd.DataFrame([{"qid": "301"}]))), 1)
        self.assertEqual(q(pd.DataFrame([{"qid": "301"}])).iloc[0].to_dict()["max-idf"], 3.5629408815)
        self.assertEqual(q(pd.DataFrame([{"qid": "301"}])).iloc[0].to_dict()["avg-idf"], 2.4740487603)

    def test_rest_query_features(self):
        import pyterrier as pt

        from tira.rest_api_client import Client
        from tira.third_party_integrations import ensure_pyterrier_is_loaded

        ensure_pyterrier_is_loaded()
        tira = Client()

        q = tira.pt.query_features(
            "ir-benchmarks/qpptk/fixed-sealer", pt.get_dataset("irds:disks45/nocr/trec-robust-2004")
        )
        self.assertEqual(len(q(pd.DataFrame([{"qid": "301"}]))), 1)
        self.assertEqual(q(pd.DataFrame([{"qid": "301"}])).iloc[0].to_dict()["features"][0], 3.5629408815)
        self.assertEqual(q(pd.DataFrame([{"qid": "301"}])).iloc[0].to_dict()["features"][1], 2.4740487603)
